/**
 * @file   eqout.c
 * @author David Bremner
 * @date   Tue Apr 16 14:50:53 2002
 * 
 * @brief  Support for outputting linear (in)equalities in various formats.
 * 
 * 
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef HAVE_CONFIG_H
#include "util_config.h"
#endif
#include "error_message.h"
#include "eqout.h"

static int *entry=NULL;
static eq_type_t  eqtype;
static eq_format_t eqform;
static int eqcols=0;
static int eqrhs;

static char *eqlabel=NULL;

static int counter=0;

extern eq_format_t eq_arg(char *optarg){
  eq_format_t outform;
  switch(optarg[0]){
  case 'i': 
    outform=INE;
    break;
  case 'c':
    outform=CPLEX;
    break;
  case 'l':
    outform=LPSOLVE;
    break;
  case 't':
    outform=TEST;
    break;
  case 'o':
    outform=OPBDP;
    break;
  default:
    err_fatal(HERE,"illegal format argument %s",optarg);
  }
  return outform;
}

extern void eq_init(eq_format_t format,int cols){
  entry=calloc(cols,sizeof(int)); 
  eqcols=cols;
  eqform=format;
}

extern void eq_zero(){
  int i;
  for (i=0; i<eqcols; i++){
    entry[i]=0;
  }
}


extern void eq_coeff(int col, int val){
  entry[col]=val;
}

extern void eq_coeff_add(int col, int val){
  entry[col]+=val;
}

extern void eq_rhs(int val){
  eqrhs=val;
}


extern void eq_set(const char *label){
  eqlabel=strdup(label);
  counter=0;
}

extern void eq_type(eq_type_t type){
  eqtype=type;
}


#define COLON_OR_NL(form) if (form==LPSOLVE) printf(":"); else printf("\n")

static void printrelation(){
    switch (eqtype){
    case LEQ:
      printf("<=");
      break;
    case GEQ:
      printf(">=");
      break;
    case EQ:
      printf("=");
      break;
    case OBJ_MAX:
    case OBJ_MIN:
      break;
    }
}

extern void eq_test_output(){
  int i;
  printrelation();
  printf("\t%d\t",eqrhs);

  for (i=0; i<eqcols; i++){
    if (entry[i]!=0){
      if (entry[i]*entry[i]!=1){
	printf("%d*",entry[i]);
      } else{
	if (entry[i]<0) {
	  printf("-");
	}
      }
      printf("%d\t",i+1);
    }
  }  
  printf("\n");
}

extern void eq_output(){

  if (eqform==TEST){
    eq_test_output();
  }
  if (eqform==CPLEX || eqform==LPSOLVE || eqform==OPBDP ){
    int i;

    if (eqform != OPBDP){
      switch(eqtype){
      case OBJ_MAX:
	printf("max");
	COLON_OR_NL(eqform);
	break;
      case OBJ_MIN:
	printf("min");
	COLON_OR_NL(eqform);
	break;
      default:
	break;
	/* do nuthin */
      }
    }
    if (eqlabel != NULL)
      printf("%s%d: ",eqlabel,counter++);

    for (i=0; i<eqcols; i++){
      if (entry[i]!=0){
	if (entry[i] > 0 ) {
	  printf("+");
	}
	if (entry[i]!=1){
	  printf("%d ",entry[i]);
	}
	printf("x%d",i);
      }
    }
    printrelation();
    if (eqtype != OBJ_MIN && eqtype !=OBJ_MAX)
      printf("%d",eqrhs);
    if (eqform == LPSOLVE)
      printf(";");
    printf("\n");
  } else {
    printf("%d",eqrhs);

    if (eqform==INE){
      int j;
      for (j=0; j<eqcols; j++){
	  printf("\t%d",-1*entry[j]);
      }
      printf("\n");
    } else {
      int j;
      for (j=0; j<eqcols; j++){
	if (entry[j]<0){
	  printf("\t%d",-1*j);
	} else if (entry[j]>0){
	  printf("\t%d",j);
	}
      }
      printf("\n");
    }
  }
}
  


