#include <stdio.h>
#include <string.h>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "error_message.h"

#include "combinat.h"
#include "oms_types.h"
#include "nalloc.h"
#include "statearray.h"
#include "nodedef.h"
#include "globaldef.h"
#include "output.h"
#include "profile.h"
#include "signed_index.h"
#include "int_to_symbol.h"
#include "nu_sign.h"
#include "memalloc.h"
#ifdef USE_ZRAM
#include <z_base.h>
#include <z_io.h>
#endif
#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#define PRINTF (void)printf


#ifdef USE_ZRAM
#define REDIR_STDOUT(buf)  z_io_RedirectStdout(buf)
#define REDIR_STDERR(buf)  z_io_RedirectStderr(buf)
#else
#define REDIR_STDOUT(buf) freopen(buf,"w",stdout);
#define REDIR_STDERR(buf) freopen(buf,"w",stderr);
#endif

static void output_chi_verbose(params_p params, sarray_p chirotope);
static void output_boundary_verbose(params_p params, sarray_p boundary);

#ifdef HAVE_GENX
#include <genx.h>

static genxWriter xml;
static  char xmlbuf[BUFSIZ];

#endif

extern void output_start_om(params_p params){


  if (strlen(params->ident) && params->output_level!=OL_NONE ){
    fprintf(stdout,"#%s\n",params->ident,prof_get(PROF_SUCCESS));
  }


  switch(params->output_level){
  case OL_COUNT:
    break;
  case OL_RLE:
  case OL_PM:
    printf("%d,%d: ",params->n,params->r);
    break;
  case OL_LIST:
  case OL_PAIRS:
  case OL_SETS:
    /* Only print the rank and number of elements for the first one*/
    if (prof_get(PROF_SUCCESS) <= 1){
      PRINTF("rank=%d\nelements=%d\n",params->r,params->n);
    }
    break;
  case OL_XML:
#ifdef HAVE_GENX
    genxStartElementLiteral(xml,NULL,"matroid-polytope");
    snprintf(xmlbuf,BUFSIZ,"%d",params->r);
    genxAddAttributeLiteral(xml,NULL,"rank",xmlbuf);
    snprintf(xmlbuf,BUFSIZ,"%d",params->n);
    genxAddAttributeLiteral(xml,NULL,"elements",xmlbuf);
    snprintf(xmlbuf,BUFSIZ,"%d",prof_get(PROF_SUCCESS));
    genxAddAttributeLiteral(xml,NULL,"count",xmlbuf);
#endif
    break;
			    
  }


}

void output_end_om(params_p params){
  switch (params->output_level){
  case OL_XML:
#ifdef HAVE_GENX
    genxEndElement(xml); /* close off the matroid polytope */
#endif
    break;
  default:
    printf("\n");
  }
}
char int_to_symbol(int s);

extern char *string_pm(sarray_p chirotope, icharfunc convert){


  int i;
  char *buf=malloc(sa_size(chirotope)+1*sizeof(char));

  for (i=0; i< sa_size(chirotope); i++){
    int s=sa_read(chirotope,i);
    buf[i]=convert(s); 
  }
  buf[i]=0; /* Null terminate */
  return buf;
}

extern void output_pm(sarray_p chirotope,icharfunc cvt){
  char *buf;
  buf=string_pm(chirotope,cvt);
  fputs(buf,stdout);
  free(buf);

}


extern void output_rle(sarray_p S,icharfunc cvt){
  int i;
  int sigma,last_sigma;
  int count=1;
  
  last_sigma=sa_read(S,0);
  for (i=1 ; i< sa_size(S); i++){
    sigma=sa_read(S,i);
    if (sigma==last_sigma){
      count++;
    }  else {
      if (count==1) 
	printf("%c",cvt(last_sigma));
      else
	printf("%d%c",count,cvt(last_sigma));
      count=1;
      last_sigma=sigma;
    }

  }

  if (count==1) 
    printf("%c",cvt(last_sigma));
  else
    printf("%d%c",count,cvt(last_sigma));

}

extern void output_chi(params_p params, sarray_p chirotope){
 
  switch(params->output_level){
  case OL_PM:
    output_pm(chirotope,sign2char);
    break;
  case OL_RLE:
    output_rle(chirotope,sign2prefix);
    break;
  case OL_COUNT:
    fprintf(stderr,"%lu found\n",prof_get(PROF_SUCCESS));
    break;
  case OL_NONE:
    break;
  default:
    output_chi_verbose(params,chirotope);
    break;
  }
}

static void output_boundary_xml(params_p params, sarray_p boundary){
  int i;

#ifdef HAVE_GENX  
  
  genxStartElementLiteral(xml,NULL,"boundary");
  for (i=0; i< sa_size(boundary); i++){
    int s;
    s=sa_read(boundary,i);
    genxStartElementLiteral(xml,NULL,"facet");
    snprintf(xmlbuf,BUFSIZ,"%d",s);
    genxAddAttributeLiteral(xml,NULL,"sign",xmlbuf);
    snprintf(xmlbuf,BUFSIZ,"%d",i+1);
    genxAddAttributeLiteral(xml,NULL,"index",xmlbuf);
    genxEndElement(xml);
  }
  genxEndElement(xml);

#endif
}

extern void output_boundary(params_p params, sarray_p boundary){
 
  if (params->boundary_output){
    switch(params->output_level){
    case OL_PM:
      PRINTF(" ; ");
      output_pm(boundary,bdy2char);
      printf("\n");
      break;
    case OL_RLE:
      PRINTF(" ; ");
      output_rle(boundary,bdy2char);
      PRINTF("\n");
      break;
    case OL_COUNT:
      break;
    case OL_NONE:
      break;
    case OL_SETS:
      output_boundary_verbose(params,boundary);
      break;
    case OL_XML:
      output_boundary_xml(params,boundary);
      break;
    }
  } else {
    /*    fputs("\n",stdout);*/
  }
}

void output_state(na_node_p global,na_node_p S){

  sarray_p chirotope,boundary;
  params_p params;

  params=na_access(global,SLOT_params);
  chirotope=na_access(S,SLOT_chirotope);
  boundary=na_access(S,SLOT_boundary);
  
  /*  na_report(S); */
  if ((prof_get(PROF_SUCCESS) % params->print_every) != 0)
    return;

  output_start_om(params);
  output_chi(params,chirotope);
  output_boundary(params,boundary);
  output_end_om(params);

}

void output_verbose(params_p params, sarray_p chirotope, sarray_p boundary){
  int i;
  
  natural_t *scratch=0;

  if (params->output_level == OL_SETS )
    scratch=calloc(params->r,sizeof(natural_t));
  

  PRINTF("instance={\n");
  if (prof_get(PROF_SUCCESS) >=1)
    PRINTF("options={\ncomment=\"solution %d\"\n}\n",prof_get(PROF_SUCCESS));
  if (params->diameter != -1){
    PRINTF("distance(%d,%d)=%d\n",
	   params->source_facet+1,
	   params->sink_facet+1,
	   params->diameter);
  }
  
  
  output_chi_verbose(params,chirotope);
  output_boundary_verbose(params,boundary);
  PRINTF("}\n"); /* close instance */
  
  PRINTF("\n");
  if (params->output_level==OL_SETS) {
    free(scratch);
  }
}





static void output_boundary_verbose(params_p params, sarray_p boundary){
  int i;
  natural_t *scratch=0;

  if (params->output_level == OL_SETS )
    scratch=calloc(params->r,sizeof(natural_t));

  if ((params->boundary_output)){
    PRINTF("\nboundary={\n");
    for (i=0; i< sa_size(boundary); i++){
      int s;
      s=sa_read(boundary,i);
    
      if (s){
	if(si_sign(s) < 0 )
	  putchar('-');
      
	if (params->output_level == OL_SETS)
	  output_bindex(params->n,params->r-1,i,scratch);
	else
	  printf("%d",i+1);
	fputs("\n",stdout);
      }
    } 
    PRINTF("}\n");
    
  } else {
    PRINTF("\n");
  } 
  if (params->output_level==OL_SETS) {
    free(scratch);
  }

}




/** 
 * Take care of all of the list output formats
 * 
 * @param params global problem parameters
 * @param chi chirotope
 */
static void output_chi_verbose(params_p params, sarray_p chirotope){
  int i;

  natural_t *scratch=0;


  if (params->output_level == OL_SETS )
    scratch=calloc(params->r,sizeof(natural_t));
  

  if (params->output_level==OL_XML){
#ifdef HAVE_GENX
    genxStartElementLiteral(xml,NULL,"chirotope");
#endif
  } else {
    PRINTF("chirotope={\n");
  }

  for (i=0; i< sa_size(chirotope); i++){
    int sigma=sa_read(chirotope,i);
    

    if (!sign_unknown(sigma)){
      if (i>0 && params->output_level!=OL_XML) printf (",\n");


      switch (params->output_level){
      case OL_LIST:
	if (sigma != SIGN_POS )
	  putchar(sign2prefix(sigma));
	printf("%d",i+1);
	break;
      case OL_PAIRS:
	printf("(%d,%s)",i+1,sign2string(sigma));
	break;
      case OL_SETS:
	putchar('(');
	output_bindex(params->n,params->r,i,scratch);
	printf(",%s)",sign2string(sigma));
	break;
      case OL_XML:
#ifdef HAVE_GENX
	genxStartElementLiteral(xml,NULL,"basis");
	genxAddAttributeLiteral(xml,NULL,"sign",sign2string(sigma));
	snprintf(xmlbuf,BUFSIZ,"%d",i+1);
	genxAddAttributeLiteral(xml,NULL,"index",xmlbuf);
	genxEndElement(xml);
#endif
	break;
      default:
	err_fatal(HERE,"unknown output level %d",params->output_level);
      }
    }
  }
  if (params->output_level==OL_XML){
#ifdef HAVE_GENX
    genxEndElement(xml);
#endif
  } else {
    PRINTF("\n}\n");
  }
  if (params->output_level==OL_SETS) {
    free(scratch);
  }

}

void oms_redir_output(char *name){


  char buf[BUFSIZ];

  snprintf(buf,BUFSIZ,"%s-out",name);
  REDIR_STDOUT(buf);
  snprintf(buf,BUFSIZ,"%s-err",name);
  REDIR_STDERR(buf);
}

void output_init(params_p params){
  if (params->output_level==OL_XML){
#ifdef HAVE_GENX  
    xml=genxNew(NULL,NULL,NULL);
    genxStartDocFile(xml,stdout);
    genxStartElementLiteral(xml,NULL,"matroid-polytope-list");
    snprintf(xmlbuf,BUFSIZ,"%d",params->r);
    genxAddAttributeLiteral(xml,NULL,"rank",xmlbuf);
    snprintf(xmlbuf,BUFSIZ,"%d",params->n);
    genxAddAttributeLiteral(xml,NULL,"elements",xmlbuf);
#else
    err_fatal(HERE,"XML Support not compiled in");
#endif
  }
}

void output_finish(params_p params){
  if (params->output_level==OL_XML){
#ifdef HAVE_GENX  
    genxEndElement(xml); /* close list */
    genxEndDocument(xml);
#else
    err_fatal(HERE,"XML Support not compiled in");
#endif
  } else {
      
    if (params->profile_level >0 ){
      fprintf(stdout,"# %lu solutions found\n", prof_get(PROF_SUCCESS));
    }
  }
}
