/* Branch and bound interface example
   Ambros Marzetta

 * $Source$
 * $Revision: 2934 $
 * $Date: 2002-12-06 06:12:57 -0400 (Fri, 06 Dec 2002) $
 * $Log$
 * Revision 1.3  2002/12/06 10:12:57  bremner
 * found one line bug in startup.c after moving heaven and earth
 * build zram examples (some)
 *
 * Revision 1.2  2000/11/15 16:14:51  bremner
 *
 * committing patched version of zram source
 *
 * Revision 1.2  1996/06/03  18:59:12  marzetta
 * New definition of z_NewNode.
 *
 * Revision 1.1  1996/05/29  18:07:51  marzetta
 * Initial revision
 *
 * Revision 1.10  1996/05/08  20:28:09  marzetta
 * Added depth-first with explicit stack, restarting
 *
 * Revision 1.9  1996/04/09  13:15:08  marzetta
 * Estimator
 *
 * Revision 1.8  1995/08/22  12:39:12  marzetta
 * New message passing, MPI
 *
 * Revision 1.7  1995/08/02  13:28:46  marzetta
 * Portability to macintosh
 *
 * Revision 1.6  1995/07/31  14:43:48  marzetta
 * Branch in place
 *
 * Revision 1.5  1995/07/26  09:32:32  marzetta
 * Branch procedure gets current upper bound
 *
 * Revision 1.4  1995/07/25  11:52:14  marzetta
 * s_bb1_Init, float32, upper bound
 *
 * Revision 1.3  1995/07/25  11:16:33  marzetta
 * Introduced int16
 *
 * Revision 1.1  1995/07/24  13:57:16  marzetta
 * Initial revision
 *
 */

#include <stdio.h>
#include "z_base.h"
#include "z_bb.h"
#include "z_io.h"		/* for saving the global data */
#include "z_signal.h"

static int example_NodeTag, example_GlobalTag; /* for z_InstallClass */

typedef struct example_Node {
  Z_NODE_HEADER;		/* len; tag = example_NodeTag */
  int16 nr;			/* reference number of this node */
  float32 lb;			/* lower bound */
} example_Node, *example_NodeP;

typedef struct example_Global {
  Z_NODE_HEADER;		/* len; tag = example_GlobalTag */
  int16 limit;			/* all generated nodes have nr < limit */
} example_Global, *example_GlobalP;

/* Static variables ------------------------------------------------ */

static z_bb_ProcSet example_procset;
static int example_procsetnum;

/* Procedures ------------------------------------------------------ */

static boolean example_IsSolution(example_NodeP np)
{
  /* Is np a solution (leaf node of branch and bound tree)? */
  assert(np && np->tag == example_NodeTag);

  return np->nr == 69 || np->nr == 38 || np->nr == 85 || np->nr == 60 ||
    np->nr == 58 || np->nr == 61;
} /* example_IsSolution */

static void example_Branch(example_GlobalP global,
		    example_NodeP np, example_NodeP upperBound,
		    z_bb_DoProcP whatToDo, void *ref)
{
  /* Allocate nodes for children of np. */
  /* Call whatToDo(child, ref) for all children of np. */
  /* whatToDo is responsible for deallocating the children. */
  /* whatToDo may call example_Branch recursively. */
  /* np must not be a solution. */
  
  example_NodeP left, right;

  assert(global && global->tag == example_GlobalTag);
  assert(np && np->tag == example_NodeTag);
  assert(! example_IsSolution(np));

  if (np->nr * 2 < global->limit) {
    left = (example_NodeP) z_NewCopyNode((z_NodeP) np);
    left->nr = np->nr * 2;
    left->lb = np->lb * 1.2;
    whatToDo((z_NodeP) left, ref);
  }
  if (np->nr * 2 + 1 < global->limit) {
    right = (example_NodeP) z_NewCopyNode((z_NodeP) np);
    right->nr = np->nr * 2 + 1;
    right->lb = np->lb * 1.1;
    whatToDo((z_NodeP) right, ref);
  }
} /* example_Branch */

static void example_BranchInPlace(example_GlobalP global,
			   example_NodeP np, example_NodeP upperBound,
			   z_bb_DoProcP whatToDo, void *ref)
{
  /* Generate children of np by modifying it. */
  /* Call whatToDo(np, ref) for all children of np. */
  /* Revert np back to its original value. */
  /* whatToDo is responsible for copying the children (if necessary). */
  /* whatToDo may call example_Branch recursively. */
  /* np must not be a solution. */
  
  float32 old_lb;

  assert(global && global->tag == example_GlobalTag);
  assert(np && np->tag == example_NodeTag);
  assert(! example_IsSolution(np));

  old_lb = np->lb;
  
  if (np->nr * 2 < global->limit) {
    np->nr = np->nr * 2;
    np->lb = old_lb * 1.2;
    whatToDo((z_NodeP) np, ref);
    np->nr /= 2;
  }
  if (np->nr * 2 + 1 < global->limit) {
    np->nr = np->nr * 2 + 1;
    np->lb = old_lb * 1.1;
    whatToDo((z_NodeP) np, ref);
    np->nr /= 2;
  }
  np->lb = old_lb;
} /* example_BranchInPlace */

static example_NodeP example_UpperBound(example_GlobalP global, example_NodeP n)
{
  assert(global && global->tag == example_GlobalTag);
  assert(n && n->tag == example_NodeTag);
  assert(! example_IsSolution(n));

  /* This is not a correct upper bound. But it shows how the procedure
     should work.
     */
  n = (example_NodeP) z_NewCopyNode((z_NodeP) n);
  n->nr = 999;
  n->lb = 100;
  return n;
} /* example_UpperBound */

static int example_CompareProc(example_NodeP x, example_NodeP y)
{
  /* Compare the lower bounds of x and y */
  assert(x && x->tag == example_NodeTag);
  assert(y && y->tag == example_NodeTag);

  if (x->lb > y->lb) return 1;
  else if (x->lb == y ->lb) return 0;
  else return -1;
} /* example_CompareProc */

static char *example_ShortNodeString(example_NodeP np)
{
  /* Return a static string describing np */
  static char nodeString[80];

  if (np) {
    assert(np->tag == example_NodeTag);
    sprintf(nodeString, "nr=%d, lb=%f",
	    (int)np->nr,
	    np->lb);
    return nodeString;
  } else
    return "NULL Node";
} /* example_ShortNodeString */

void Estimate(example_GlobalP global, example_NodeP root)
{
  /* Estimate the number of nodes in the search tree */
  example_NodeP ub;
  float nNodes, sNodes, nTime, sTime;

  printf("***** Estimating the number of nodes\n");
  ub = (example_NodeP) z_NewNode(example_NodeTag, sizeof(example_Node));
  ub->nr = 0;
  ub->lb = 2.0;

  z_BranchBoundEstimate(example_procsetnum,
			(z_NodeP)global,
			(z_NodeP)root,
			(z_NodeP)ub,
			1000,
			&nNodes, &sNodes,
			&nTime, &sTime);

  printf("Estimate: %5.1f +- %5.1f nodes in search tree\n", nNodes, sNodes);
  printf("Estimated time: %5.4f +- %5.4fs\n", nTime, sTime);
  
  z_FreeNode((z_NodeP)ub);
} /* Estimate */

void SolveSequentially(example_GlobalP global, example_NodeP root)
{
  example_NodeP sol;

  printf("\n***** Solving sequentially\n");
  sol = (example_NodeP)
    z_BranchBoundDepthFirst(example_procsetnum,
			    (z_NodeP)global,
			    (z_NodeP)root,
			    (z_NodeP)example_UpperBound(global,
							root),
			    1);
  printf("Solution: %s\n", example_ShortNodeString(sol));
  z_FreeNode((z_NodeP)sol);
} /* SolveSequentially */

const char globalFileName[] = "gf"; /* should be read from command line */

int Restart(void)
{
  FILE *gf;
  example_GlobalP global;
  example_NodeP sol;

  printf("***** Restarting from checkpoint\n");
  gf = fopen(globalFileName, "r");
  if (!gf) {
    perror(globalFileName);
    return 9;
  }
  global = (example_GlobalP) z_ReadNode(gf);
  fclose(gf);
  z_StartCheckpointTimer(3600);
  sol = (example_NodeP) z_RestartBranchBoundDFExp(example_procsetnum,
						  (z_bb_GlobalDataP) global,
						  0);
  printf("Solution: %s\n", example_ShortNodeString(sol));
  return 0;
} /* Restart */

void initproc(void)
{
  /* This function is called once on every processing element. */

  z_signal_Init();
  s_bb1_Init();
  z_bb1_Init();
  z_bb2_Init();
  z_bb3_Init();

  example_NodeTag = z_InstallClass(example_Node, 0, 0, 0);
  example_GlobalTag = z_InstallClass(example_Global, 0, 0, 0);
  example_procset.bb_Branch = (z_bb_BranchProc)example_Branch;
  example_procset.bb_BranchInPlace = (z_bb_BranchProc)example_BranchInPlace;
  example_procset.bb_UpperBound = (z_bb_UpperBoundProc) example_UpperBound;
  example_procset.bb_CompareProc = (z_bb_CompareProc)example_CompareProc;
  example_procset.bb_IsSolution = (z_bb_IsSolutionProc)example_IsSolution;
  example_procset.bb_ShortNodeString =
    (z_bb_ShortNodeStringProc)example_ShortNodeString;

  example_procsetnum = z_InstallProcSet((z_ProcSetP) &example_procset);
  setvbuf(stdout, NULL, _IONBF, 0);
} /* initproc */

int mmain(int argc, char **argv, char **envp)
{
  /* main program */
  example_GlobalP global;
  example_NodeP root;

  if (argc > 1)
    return Restart();

  global = (example_GlobalP) z_NewNode(example_GlobalTag,
				       sizeof(example_Global));
  global->limit = 128;

  root = (example_NodeP) z_NewNode(example_NodeTag, sizeof(example_Node));
  root->nr = 1;
  root->lb = 1.0;

  Estimate(global, root);
  printf("Allocated nodes: %ld\n", z_numNodes[example_NodeTag]);
  /* should be 1 (root) */

  SolveSequentially(global, root);
  printf("Allocated nodes: %ld\n", z_numNodes[example_NodeTag]);
  /* should be 1 (root) */

#if 1 /* parallel, mixed best and depth first, no checkpointing */
#ifndef SEQUENTIAL
  printf("\n***** Solving in parallel, not restartable\n");
  {
    example_NodeP sol;
    sol = (example_NodeP) z_BranchBoundParallel(example_procsetnum,
						(z_NodeP)global,
						(z_NodeP)root, NULL, 10, 1);
    printf("Solution: %s\n", example_ShortNodeString(sol));
    z_FreeNode((z_NodeP)sol);
  }
  printf("Allocated nodes: %ld\n", z_numNodes[example_NodeTag]);
  /* should be 1 (root) */
#endif /* SEQUENTIAL */
#else /* parallel, depth first, restartable */
  printf("\n***** Solving in parallel, restartable\n");
  {
    FILE *gf;
      
    gf = fopen(globalFileName, "w");	/* save for restart */
    if (!gf) {
      perror("open/w");
      return 9;
    }
    z_WriteNode(gf, (z_NodeP) global);
    fclose(gf);
  }
  z_StartCheckpointTimer(3600);
  {
    example_NodeP sol;
    sol = (example_NodeP) z_BranchBoundDFExp(example_procsetnum,
					     (z_NodeP)global,
					     (z_NodeP)root, NULL, 1, 0);
    printf("Solution: %s\n", example_ShortNodeString(sol));
    z_FreeNode((z_NodeP)sol);
  }
  printf("Allocated nodes: %ld\n", z_numNodes[example_NodeTag]);
  /* should be 1 (root) */
#endif

  return 0;
} /* mmain */
