/* s_bb1.c (ZRAM: Branch and Bound)
   Ambros Marzetta

 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 12:14:52 -0400 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.11  1997/04/25  08:41:11  marzetta
 * Added some type casts.
 *
 * Revision 1.10  1997/01/08  17:20:31  marzetta
 * z_BranchBoundBestFirst: deallocate priority queue
 *
 * Revision 1.9  1996/12/03  14:04:42  marzetta
 * Fixed bug with numNodes and initial upper bound
 *
 * Revision 1.8  1996/04/09  13:34:31  marzetta
 * removed z_stdio.h
 *
 * Revision 1.7  1995/08/24  13:04:52  marzetta
 * Added some assert() calls.
 *
 * Revision 1.6  1995/08/22  12:39:12  marzetta
 * New message passing, MPI
 *
 * Revision 1.5  1995/08/02  13:28:46  marzetta
 * Portability to macintosh
 *
 */

#include <stdio.h>
#include <stdlib.h>		/* exit */
#include "z_base.h"
#include "z_bb.h"
#include "s_pq.h"
#include "z_time.h"		/* front end only */

static boolean s_bb1_inited;

#define ShortNodeString(x) glob->psp->bb_ShortNodeString ? \
  glob->psp->bb_ShortNodeString(x) : ""
  
#define ShortNodeString2(x) glob.psp->bb_ShortNodeString ? \
  glob.psp->bb_ShortNodeString(x) : ""
  
/* Branch and bound depth first ----------------------------------------- */

typedef struct DFglobalsT {
  z_bb_ProcSetP psp;
  z_NodeP solNode;
  z_bb_GlobalDataP globalData;
  long branchCount;
} DFglobalsT;

static void z_BBDF(DFglobalsT *glob, z_NodeP root);

static void DoDFChild(z_NodeP child, DFglobalsT *glob)
{
  assert(child);
  z_BBDF(glob, child);
  z_FreeNode(child);
}

static void DoDFChildInPlace(z_NodeP child, DFglobalsT *glob)
{
  assert(child);
  z_BBDF(glob, child);
} /* DoDFChildInPlace */

static void z_BBDF(DFglobalsT *glob, z_NodeP root)
{
  /* If there is a better solution in the given subtree, adjust
     solNode. Deallocate the old contents of solNode.
     Allocate the new contents.
     */

  #ifdef TRACENODE
    printf("  Node: (%s)\n", ShortNodeString(root));
  #endif
  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(root, glob->solNode) < 0) {
    if (glob->psp->bb_IsSolution(root)) {	/* leaf -> new upper bound */
      printf("New upper bound: %s\n", ShortNodeString(root));
      if (glob->solNode) z_FreeNode(glob->solNode);
      glob->solNode = z_NewCopyNode(root);
    } else {			/* not a leaf -> branch */
      glob->branchCount++;
      if (glob->psp->bb_BranchInPlace)
	glob->psp->bb_BranchInPlace(glob->globalData,
				    root, glob->solNode,
				    (z_bb_DoProcP) DoDFChildInPlace, glob);
      else
	glob->psp->bb_Branch(glob->globalData,
			     root, glob->solNode,
			     (z_bb_DoProcP) DoDFChild, glob);
    }
  } /* lb < *upperBound */
} /* z_BBDF */

z_NodeP z_BranchBoundDepthFirst(z_ProcSetNum procsetnum,
				z_bb_GlobalDataP global,
				z_NodeP root, z_NodeP initialUpperBound,
				int noisy)
{
  /* Allocate and return a solution node with minimal value */
  DFglobalsT glob;

  assert(s_bb1_inited);
  assert(root);
  glob.psp = (z_bb_ProcSetP) z_GetProcSet(procsetnum);
  assert(glob.psp->bb_Branch || glob.psp->bb_BranchInPlace);
  assert(glob.psp->bb_CompareProc);
  assert(glob.psp->bb_IsSolution);
  glob.globalData = global;
  glob.solNode = initialUpperBound;
  if (! glob.solNode)
    if (glob.psp->bb_UpperBound)
      glob.solNode = glob.psp->bb_UpperBound(glob.globalData,root);
    else
      glob.solNode = NULL;
  glob.branchCount = 0;
  if (noisy) printf("Initial upper Bound (%s)\n",
		    ShortNodeString2(glob.solNode));
  z_BBDF(&glob, root);
  if (noisy) printf("Branch: %ld\n", glob.branchCount);
  return glob.solNode;
} /* z_BranchBoundDepthFirst */


/* Branch and bound best first ------------------------------------------ */

typedef struct BFglobalsT {
  z_bb_ProcSetP psp;
  z_NodeP solNode;
  z_bb_GlobalDataP globalData;
  z_pqP pq;
  long branchCount;
  #ifdef TRACENODE
    long lastBranchCountB;
  #endif
  int noisy;
} BFglobalsT;

static void DoBFChild(z_NodeP child, BFglobalsT *glob)
{
  /* This procedure is called for all children of the node just
     extracted from the priority queue. The child either goes to
     glob->solNode, to the priority queue, or is deallocated.
     */
  assert(child);
  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(child, glob->solNode) < 0) {
				/* better than upper bound */
    if (glob->psp->bb_IsSolution(child)) { /* solution -> new upper bound*/
      if (glob->solNode) z_FreeNode(glob->solNode);
      glob->solNode = child;
      printf("New upper bound: (%s)\n",
	     ShortNodeString(child));
      z_pq_DeleteAbove(glob->pq, child, (whatToDoProcP) z_FreeNode);
    } else
      z_pq_Insert(glob->pq, child); /* no solution -> queue */
    /*printf("Insert (%s)\n", ShortNodeString(child));*/
  }
  else
    z_FreeNode(child); /* not better than upper bound -> Free */
} /* DoBFChild */

static void DoBFChildInPlace(z_NodeP child, BFglobalsT *glob)
{
  /* This procedure is called for all children of the node just
     extracted from the priority queue. The child either goes to
     glob->solNode, to the priority queue, or is deallocated.
     */
  assert(child);
  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(child, glob->solNode) < 0) {
				/* better than upper bound */
    if (glob->psp->bb_IsSolution(child)) { /* solution -> new upper bound*/
      if (glob->solNode) z_FreeNode(glob->solNode);
      glob->solNode = z_NewCopyNode(child);
      printf("New upper bound: (%s)\n",
	     ShortNodeString(child));
      z_pq_DeleteAbove(glob->pq, child, (whatToDoProcP) z_FreeNode);
    } else
      z_pq_Insert(glob->pq, z_NewCopyNode(child)); /* no solution -> queue */
    /*printf("Insert (%s)\n", ShortNodeString(child));*/
  }
} /* DoBFChildInPlace */

static void z_BBBF(BFglobalsT *glob)
{
  /* If there is a better solution in the given subtree, adjust
     solNode. Deallocate the old contents of solNode.
     Allocate the new contents.
     */
  z_NodeP node;
  z_pq_refP ref;

  for (;;) {
    if (z_pq_isEmpty(glob->pq)) break; /* no more work left */
    ref = z_pq_Minimum(glob->pq);
    node = (z_NodeP) z_pq_GetObject(glob->pq, ref);
    z_pq_Delete(glob->pq, ref);
    #ifdef TRACENODE
      if (glob->branchCount - glob->lastBranchCountB >= TRACENODE) {
	glob->lastBranchCountB = glob->branchCount;
	if (glob->noisy)
	  printf("  Node: (%s)\n",
		 ShortNodeString(node));
      }
    #endif
    assert(! glob->psp->bb_IsSolution(node));
    /* not a leaf -> branch */
    glob->branchCount++;
    if (glob->psp->bb_Branch)
      glob->psp->bb_Branch(glob->globalData,
			   node, glob->solNode,
			   (z_bb_DoProcP) DoBFChild, glob);
    else
      glob->psp->bb_BranchInPlace(glob->globalData,
				  node, glob->solNode,
				  (z_bb_DoProcP) DoBFChildInPlace, glob);
    z_FreeNode(node);
  } /* for */
} /* z_BBBF */

z_NodeP z_BranchBoundBestFirst(z_ProcSetNum procsetnum,
			       z_bb_GlobalDataP global,
			       z_NodeP root,
			       z_NodeP initialUpperBound,
			       int noisy)
{
  /* Allocate and return a solution node with minimal value */
  struct BFglobalsT glob;

  assert(s_bb1_inited);
  assert(root);
  glob.psp = (z_bb_ProcSetP) z_GetProcSet(procsetnum);
  assert(glob.psp->bb_Branch || glob.psp->bb_BranchInPlace);
  assert(glob.psp->bb_CompareProc);
  assert(glob.psp->bb_IsSolution);
  if (glob.psp->bb_IsSolution(root))
    return z_NewCopyNode(root);
  
  glob.globalData = global;
  glob.solNode = initialUpperBound;
  if (! glob.solNode)
    if (glob.psp->bb_UpperBound)
      glob.solNode = glob.psp->bb_UpperBound(global, root);
    else
      glob.solNode = NULL;
  glob.pq = z_pq_NewPriorityQueue(z_pq_default,
				  (z_pq_CompProcP)glob.psp->bb_CompareProc);
  (void) z_pq_Insert(glob.pq, z_NewCopyNode(root));
  #ifdef TRACENODE
    glob.lastBranchCountB = 0;
  #endif
  glob.noisy = noisy;
  glob.branchCount = 0;
  if (noisy)
    printf("Initial upper Bound (%s)\n",
	   ShortNodeString2(glob.solNode));
  z_BBBF(&glob);
  assert(z_pq_isEmpty(glob.pq));
  if (noisy) {
    printf("%s\n", z_pq_InfoString(glob.pq));
    printf("Branch: %ld\n", glob.branchCount);
  }
  z_pq_FreePriorityQueue(glob.pq);
  return glob.solNode;
}


/* Branch and bound mixed best and depth first ---------------------------- */

typedef struct MFglobalsT {
  z_bb_ProcSetP psp;
  z_NodeP solNode;
  z_bb_GlobalDataP globalData;
  z_pqP pq;
  int queueSize;
  long branchCount;
  #ifdef TRACENODE
    long lastBranchCountD, lastBranchCountB;
  #endif
  double lastTraceD, lastTraceB;
  int noisy;
  int tag;
  long numNodesStart;		/* nodes at start + solNode */
} MFglobalsT;

static void DoMFChildInPlace(z_NodeP child, MFglobalsT *glob);

static void DoMFChild(z_NodeP child, MFglobalsT *glob)
{
  /* This procedure is called for all children. The child either goes to
     glob->solNode, to the priority queue, or is deallocated.
     */
  assert(child);
  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(child, glob->solNode) < 0) {
    /* better than upper bound */
    if (glob->psp->bb_IsSolution(child)) { /* solution -> new upper bound*/
      if (glob->solNode)
	z_FreeNode(glob->solNode);
      else
	glob->numNodesStart ++;
      glob->solNode = child;
      if (glob->noisy)
	printf("New upper bound: (%s)\n",
	       ShortNodeString(child));
      z_pq_DeleteAbove(glob->pq, child,
		       (whatToDoProcP) z_FreeNode);
    } else { /* better than upper bound, but no solution */
      if (z_pq_numElements(glob->pq) < glob->queueSize) { /* insert */
	z_pq_Insert(glob->pq, child); /* no solution -> queue */
	/*printf("    Insert (%s)\n",
	  ShortNodeString(child));*/
      } else { /* continue depth first */
#ifdef TRACENODE
	if (glob->branchCount - glob->lastBranchCountD >= TRACENODE) {
	  glob->lastBranchCountD = glob->branchCount;
	  printf("  Node (depth first): (%s)\n",
		 ShortNodeString(child));
	  fflush(stdout);
	}
#endif
	if (glob->noisy && z_Wtime() - glob->lastTraceD >= 30) {
	  printf("  Node (depth first): (%s)\n",
		 ShortNodeString(child));
	  fflush(stdout);
	  glob->lastTraceD = z_Wtime();
	}
	glob->branchCount++;
	if (glob->psp->bb_BranchInPlace)
	  glob->psp->bb_BranchInPlace(glob->globalData,
			       child, glob->solNode,
			       (z_bb_DoProcP) DoMFChildInPlace, glob);
	else
	  glob->psp->bb_Branch(glob->globalData,
			       child, glob->solNode,
			       (z_bb_DoProcP) DoMFChild, glob);
	z_FreeNode(child);
      } /* continue depth first */
    } /* better than upper bound, but no solution */
  } /* better than upper bound */
  else {
    z_FreeNode(child); /* not better than upper bound -> Free */
  }
} /* DoMFChild */

static void DoMFChildInPlace(z_NodeP child, MFglobalsT *glob)
{
  /* This procedure is called for all children. The child either goes to
     glob->solNode, to the priority queue, or is deallocated.
     */
  assert(child);
  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(child, glob->solNode) < 0) {
    /* better than upper bound */
    if (glob->psp->bb_IsSolution(child)) { /* solution -> new upper bound*/
      if (glob->solNode)
	z_FreeNode(glob->solNode);
      else
	glob->numNodesStart ++;
      glob->solNode = z_NewCopyNode(child);
      if (glob->noisy)
	printf("New upper bound: (%s)\n",
	       ShortNodeString(child));
      z_pq_DeleteAbove(glob->pq, child,
		       (whatToDoProcP) z_FreeNode);
    } else { /* better than upper bound, but no solution */
      if (z_pq_numElements(glob->pq) < glob->queueSize) { /* insert */
	z_pq_Insert(glob->pq, z_NewCopyNode(child)); /* no solution -> queue */
      } else { /* continue depth first */
#ifdef TRACENODE
	if (glob->branchCount - glob->lastBranchCountD >= TRACENODE) {
	  glob->lastBranchCountD = glob->branchCount;
	  printf("  Node (depth first): (%s)\n",
		 ShortNodeString(child));
	  fflush(stdout);
	}
#endif
	if (glob->noisy && z_Wtime() - glob->lastTraceD >= 30) {
	  printf("  Node (depth first): (%s)\n",
		 ShortNodeString(child));
	  fflush(stdout);
	  glob->lastTraceD = z_Wtime();
	}
	glob->branchCount++;
	if (glob->psp->bb_BranchInPlace)
	  glob->psp->bb_BranchInPlace(glob->globalData,
			       child, glob->solNode,
			       (z_bb_DoProcP) DoMFChildInPlace, glob);
	else
	  glob->psp->bb_Branch(glob->globalData,
			       child, glob->solNode,
			       (z_bb_DoProcP) DoMFChild, glob);
      } /* continue depth first */
    } /* better than upper bound, but no solution */
  } /* better than upper bound */
} /* DoMFChildInPlace */

static void z_BBMF(MFglobalsT *glob)
{
  /* If there is a better solution in the given subtree, adjust
     solNode. Deallocate the old contents of solNode.
     Allocate the new contents.
     */

  z_NodeP node;
  z_pq_refP ref;

  for (;;) {
    assert(glob->numNodesStart ==
	z_numNodes[glob->tag] - z_pq_numElements(glob->pq));
    if (z_pq_isEmpty(glob->pq)) break; /* no more work left */
    ref = z_pq_Minimum(glob->pq);
    node = (z_NodeP) z_pq_GetObject(glob->pq, ref);
    z_pq_Delete(glob->pq, ref);
    assert(! glob->psp->bb_IsSolution(node));
    /* not a leaf -> branch */
    glob->branchCount++;
    #ifdef TRACENODE
      if (glob->branchCount - glob->lastBranchCountB >= TRACENODE) {
	glob->lastBranchCountB = glob->branchCount;
	if (glob->noisy) {
	  printf("  Node (queue=%5ld): (%s)\n", z_pq_numElements(glob->pq),
		 ShortNodeString(node));
	  fflush(stdout);
	}
      }
    #endif
	if (glob->noisy && z_Wtime() - glob->lastTraceB >= 30) {
	  printf("  Node (queue=%5ld): (%s)\n", z_pq_numElements(glob->pq),
		 ShortNodeString(node));
	  fflush(stdout);
	  glob->lastTraceB = z_Wtime();
	}
    if (glob->psp->bb_Branch)
      glob->psp->bb_Branch(glob->globalData,
			   node, glob->solNode,
			   (z_bb_DoProcP) DoMFChild, glob);
    else
      glob->psp->bb_BranchInPlace(glob->globalData,
				  node, glob->solNode,
				  (z_bb_DoProcP) DoMFChildInPlace, glob);
    z_FreeNode(node);
  }
} /* z_BBMF */

z_NodeP z_BranchBoundMixedFirst(z_ProcSetNum procsetnum,
				z_bb_GlobalDataP global,
				z_NodeP root, z_NodeP initialUpperBound,
				int queueSize, int noisy)
{
  /* Allocate and return a solution node with minimal value */
  MFglobalsT glob;

  assert(s_bb1_inited);
  assert(root);
  glob.psp = (z_bb_ProcSetP) z_GetProcSet(procsetnum);
  assert(glob.psp->bb_Branch || glob.psp->bb_BranchInPlace);
  assert(glob.psp->bb_CompareProc);
  assert(glob.psp->bb_IsSolution);
  if (glob.psp->bb_IsSolution(root))
    return z_NewCopyNode(root);

  glob.globalData = global;
  glob.tag = root->tag;
  glob.numNodesStart = z_numNodes[glob.tag];
  
  glob.solNode = initialUpperBound;
  if (! glob.solNode)
    if (glob.psp->bb_UpperBound) {
      glob.solNode = glob.psp->bb_UpperBound(glob.globalData,root);
      if (glob.solNode) glob.numNodesStart ++;
    } else
      glob.solNode = NULL;
  glob.pq = z_pq_NewPriorityQueue(z_pq_default,
				  (z_pq_CompProcP)glob.psp->bb_CompareProc);
  (void) z_pq_Insert(glob.pq, z_NewCopyNode(root));
  glob.queueSize = queueSize;
  glob.branchCount = 0;
  #ifdef TRACENODE
    glob.lastBranchCountD = 0;
    glob.lastBranchCountB = 0;
  #endif
  glob.lastTraceD = glob.lastTraceB = z_Wtime();
  glob.noisy = noisy;
  if (noisy)
    printf("Initial upper Bound (%s)\n",
	   ShortNodeString2(glob.solNode));
  z_BBMF(&glob);
  if (noisy) {
    printf("%s\n", z_pq_InfoString(glob.pq));
    printf("Branch: %ld\n", glob.branchCount);
  }
  assert(z_pq_isEmpty(glob.pq));
  z_pq_FreePriorityQueue(glob.pq);
  assert(glob.numNodesStart == z_numNodes[glob.tag]);
  return glob.solNode;
} /* z_BranchBoundMixedFirst */

/* Initialisation ------------------------------------------------------- */

void s_bb1_Init(void)
{
  if (! s_bb1_inited) {
    s_bb1_inited = 1;
  }
} /* s_bb1_Init */
