/* z_spq.c (speculative priority queues)

   Ambros Marzetta

   95-05-17

 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 12:14:52 -0400 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.5  1996/04/09  13:23:25  marzetta
 * Added SEQUENTIAL and some comments
 *
 * Revision 1.4  1995/08/22  12:39:45  marzetta
 * New message passing, MPI
 *
 * Revision 1.3  1995/08/02  08:36:11  marzetta
 * Implemented z_spq_Free, corrected GotEmpty
 *
 */

#include <stdlib.h> /* malloc */
#include <stdio.h> /* sprintf for z_pq_InfoString */
#include <string.h> /* memcpy */
#include <assert.h>
#include "z_par.h"
#include "z_mp.h"
#include "z_spq.h"

#ifndef SEQUENTIAL

#define TRACESENDTO0

/* Types ----------------------------------------------------------- */

struct z_spqT {
  z_pqP localQueue;
  int startup0;
  int startup;			/* > 0 during startup phase */
  long next_balance;		/* when to send the next nodes away */
  int next_balance_dir;		/* in which direction to send it */
  int has[MAX_PE];		/* how many nodes the other PEs have */
  int numhas;			/* number of entries != -1 in has[] */
  long numIns, numDel, numDelAbove, numMin, maxElem; /* for statistics */
};

#define MAXQUEUES 10
static struct z_spqT spq[MAXQUEUES];
static int numqueues;

static long someNodeTag, emptyTag;
static z_spq_inited;

/* Procedures (general) --------------------------------------------- */

static void SendNodesTo(z_spqP glob, int dest, int num)
{
  /* Send num nodes to dest. */
  z_NodeP some;
  z_pq_refP ref;
  int i;

  if (num > 0) {
    long maxnum;
    ref = z_pq_Minimum(glob->localQueue);
    some = z_pq_GetObject(glob->localQueue, ref);
    maxnum = (z_bufsize - 5 * sizeof(MINT)) / some->len;
    /* maxnum > 0: otherwise the first broadcast would have failed */
    if (num > maxnum) num = maxnum;
  }
  
  z_AddTagBuffer(someNodeTag);
  z_AddMINTBuffer(z_myproc);	/* my PE number */
  z_AddMINTBuffer(glob - spq);	/* my queue number */
  z_AddMINTBuffer(z_pq_numElements(glob->localQueue) - num);
				/* number of nodes in my local queue */
  z_AddMINTBuffer(num);		/* number of nodes sent */

  for (i = num; i > 0; i--) {
    ref = z_pq_Minimum(glob->localQueue);
    some = z_pq_GetObject(glob->localQueue, ref);
    z_pq_Delete(glob->localQueue, ref);
    z_AddNodeBuffer(some);
    z_FreeNode(some);
  }
  
  z_Isend(dest, "some nodes");
#ifdef TRACESENDTO
  printf("[%2d]Send to PE %d, num=%d, I have %d\r\n",
	 z_myproc, dest, num, z_pq_numElements(glob->localQueue));
#endif
} /* SendNodesTo */

static void SendNodes(z_spqP glob, int num)
{
  /* Send num nodes to some other PE. Modify next_balance_dir.
   Pre: CanSend */

  SendNodesTo(glob, z_myproc ^ glob->next_balance_dir, num);

  glob->next_balance_dir *= 2;
  while ((z_myproc ^ glob->next_balance_dir) >= z_numprocs) {
    glob->next_balance_dir *= 2;
    if (glob->next_balance_dir >= z_numprocs) {
      glob->startup /= 2;
      glob->next_balance_dir = 1;
    }
  }
} /* SendNodes */

static void PollSend(z_spqP this_spq)
{
  /* Send a message if necessary and possible */
  if (this_spq->next_balance > 0)
    this_spq->next_balance --;
  assert(this_spq->next_balance >= 0);
  
  if (z_numprocs > 1 && z_TestSend()) {
    if (!this_spq->startup && this_spq->numhas > 0) {
				/* 1st priority: give where necessary */
				/* second message of message pair */
      while ((this_spq->numhas > 0) &&
	     (! z_pq_isEmpty(this_spq->localQueue)) &&
	     z_TestSend()) {
	int i, diff;
	
	/*printf("[%2d]%3d:", z_myproc, this_spq->numhas);*/
	/*for (i = z_numprocs - 1; i >= 0; i--) printf("%3d", this_spq->has[i]);*/
	/*printf("\r\n");*/
	for (i = z_numprocs - 1; i >= 0 && (this_spq->has[i] == -1); i--) ;
	assert(i >= 0);
	diff = (z_pq_numElements(this_spq->localQueue) - this_spq->has[i]);
#ifdef TRACEHAS
	printf("[%2d]Balance: %2d has%3d, I have%3d\r\n",
	       z_myproc, i, this_spq->has[i], this_spq->has[i] + diff);
#endif
	if (diff > 5)
	  SendNodesTo(this_spq, i, diff / 5);
	this_spq->has[i] = -1;
	this_spq->numhas --;
	/*printf("[%2d]---%d %d\r\n", z_myproc, this_spq->numhas, this_spq->has[i]);*/
      } /* while */
    } else {			/* first message of message pair */
      while ((this_spq->startup || this_spq->next_balance == 0) &&
	     (! z_pq_isEmpty(this_spq->localQueue)) && z_TestSend()) {
	if (z_pq_numElements(this_spq->localQueue) > 100)
	  this_spq->next_balance = 60;
	else
	  this_spq->next_balance = 10;
	
	if (z_pq_numElements(this_spq->localQueue) > 100)
	  SendNodes(this_spq, 10);
	else if (z_pq_numElements(this_spq->localQueue) > 20)
	  SendNodes(this_spq, 10);
	else if (z_pq_numElements(this_spq->localQueue) > 5)
	  SendNodes(this_spq, 2);
	else
	  SendNodes(this_spq, 1);
      } /* while */
    }
  } /* if z_TestSend() */
} /* PollSend */

z_spqP z_spq_NewSpeculativePriorityQueue(z_pq_CompProcP compare)
/* Allocate an empty speculative priority queue. This procedure must be
   called on all processors at the same time. */
{
  z_spqP this;

  assert(z_spq_inited);
  assert(numqueues < MAXQUEUES);
  assert(! numqueues);		/* more than one spq is not yet implemented */
				/* how does termination detection for */
				/* more than one spq work? */
  this = & spq[numqueues++];
  this->localQueue = z_pq_NewPriorityQueue(z_pq_default, compare);
  this->next_balance = 1;
  this->next_balance_dir = 1;
  if (z_numprocs > 1) {
    while (this->next_balance_dir <= z_myproc) this->next_balance_dir *= 2;
    while ((z_myproc ^ this->next_balance_dir) >= z_numprocs) {
      this->next_balance_dir *= 2;
      if (this->next_balance_dir >= z_numprocs)
	this->next_balance_dir = 1;
      /*printf("[%2d]dir = %d %d %d %d\r\n", z_myproc, this->next_balance_dir,
	     z_myproc ^ this->next_balance_dir, z_numprocs,
	     ((z_myproc ^ this->next_balance_dir) >= z_numprocs));*/
    }
  }
  this->startup = 1;
  this->startup0 = 1;
  {
    int i;

    for (i = z_numprocs-1; i>= 0; i--) this->has[i] = -1;
    this->numhas = 0;
  }
  return this;
} /* z_spq_NewSpeculativePriorityQueue */

void z_spq_FreeSpeculativePriorityQueue(z_spqP pq)
/* Deallocate the speculative priority queue pq, which must be empty.
   This procedure must be called on all processors at the same time. */
{
  assert(pq == & spq[numqueues-1]); /* otherwise not yet implemented */
  z_pq_FreePriorityQueue(pq->localQueue);
  numqueues --;
} /* z_spq_FreeSpeculativePriorityQueue */

void z_spq_Insert(z_spqP pq, z_NodeP x)
/* Insert x into the speculative priority queue. */
{
  /*printf("[%2d]insert\n", z_myproc);*/
  (void) z_pq_Insert(pq->localQueue, x);
  pq->startup0 = 0;
}

static void SendEmptyMessage(z_spqP this)
{
  int dir, dest;
  
  if (!this->startup0 && (z_numprocs > 1) && z_TestSend()) {
    dir = z_numprocs - 1;	/* must be positive */
    dest = (z_myproc + dir) % z_numprocs;
    z_AddTagBuffer(emptyTag);
    z_AddMINTBuffer(z_myproc);
    z_AddMINTBuffer(this - spq);
    z_AddMINTBuffer(dir);
    z_Isend(dest, "empty (start)");
  } /* z_TestSend */
} /* SendEmptyMessage */

z_NodeP z_spq_GetDeleteNearMinimum(z_spqP pq)
/* Retrieve an element near the minimum and delete it. Return NULL if the
   speculative queue is empty.
   Pre: receive pending
   Post: receive pending iff result != NULL */
{
  boolean terminate;
  z_pq_refP ref;
  z_NodeP node;
  
  z_Poll();

  if (z_pq_isEmpty(pq->localQueue)) /* local queue is empty */
    SendEmptyMessage(pq);
  
  node = NULL;
  terminate = 0;
  while (!terminate) {
    /* receive pending */

    if (!z_pq_isEmpty(pq->localQueue)) { /* local queue is not empty */

      /*printf("[%2d]delete\n", z_myproc);*/
      ref = z_pq_Minimum(pq->localQueue);
      node = z_pq_GetObject(pq->localQueue, ref);
      z_pq_Delete(pq->localQueue, ref);
      PollSend(pq);
      terminate = 1;
      /* receive pending */

    } else { /* local queue is empty */

      /*PollSend(pq);*/
      /*printf("[%2d]local queue is empty\n", z_myproc);*/
      terminate = z_IdleWithHandler(0);
      /* receive pending iff !terminate */
    } /* local queue is empty */
  } /* while (!terminate) */
  /* receive pending iff result != NULL */
  return node;
} /* z_spq_GetDeleteNearMinimum */

long z_spq_numLocalElements(z_spqP pq)
/* Return the number of elements in the local part of pq. */
{
  return z_pq_numElements(pq->localQueue);
}

void z_spq_DeleteLocalAbove(z_spqP pq, z_NodeP upperBound,
			    whatToDoProcP whatToDo)
/* Remove all elements >= upperBound from the local priority queue. Call
   whatToDo for each of them. WhatToDo must not access the priority
   queue, but should e.g. deallocate the elements. */
{
  z_pq_DeleteAbove(pq->localQueue, upperBound, whatToDo);
}

char *z_spq_InfoString(z_spqP pq)
{
  /* Return statistical information about the queue. */
  static char s[80];

  sprintf(s, "********");
  return s;
}

static void GotEmpty(void)
{
  /* Message handler: another processor has sent an empty message
     because its local priority queue is empty.
     Do one of the following
     - Send nodes back (if I have enough nodes)
     - forget it (if the message is from myself and has been forwarded by all)
     - forward the empty message
     - forget it (if I can not send immediately)
     */
  int src, que, dir, num, dest;
  z_spqP this;

  src = z_GetMINTBuffer();	/* PE whose local queue is empty */
  que = z_GetMINTBuffer();	/* queue number */
  dir = z_GetMINTBuffer();	/* direction to send this message */
  this = &spq[que];

  if (src != z_myproc) {	/* message is not from myself */
    if (z_TestSend()) {
      if (z_spq_numLocalElements(this) >= 2) { /* send nodes back */
	if (z_spq_numLocalElements(this) >= 7)
	  num = 3;
	else
	  num = 1;
	SendNodesTo(this, src, num);
      } else {			/* forward the empty message */
	dest = (z_myproc + dir) % z_numprocs;
	z_AddTagBuffer(emptyTag);
	z_AddMINTBuffer(src);
	z_AddMINTBuffer(que);
	z_AddMINTBuffer(dir);
	z_Isend(dest, "empty");
      }
    } /* z_TestSend */
  } /* message is not from myself */
} /* GotEmpty */

static void GotSomeNodes(void)
     /* Message handler */
{
  int num, i, has, src, que;
  z_NodeP node;
  z_spqP this;
  
  src = z_GetMINTBuffer();	/* PE number of source */
  que = z_GetMINTBuffer();	/* queue number */
  has = z_GetMINTBuffer();	/* number of nodes in his queue */
  num = z_GetMINTBuffer();	/* number of nodes sent */
#ifdef TRACESOMENODE
  printf("[%2d]<Some node from %d, he has %d, I have %d\r\n",
	 z_myproc, src, has, z_pq_numElements(glob->pq));
#endif
  this = &spq[que];
  for (i = num; i > 0; i-- ) {
    node = z_GetNodeBuffer();
    /* should compare every received node to upper bound,
     but the speculative priority queue does not know the upper bound */
    (void) z_spq_Insert(this, node);
  }
  assert(has >= 0);
  if (has < 100 && z_spq_numLocalElements(this) - has >= 5) {
    if (this->has[src] == -1) this->numhas ++;
    this->has[src] = has;
    /*printf("[%2d]++%d %d %2d\r\n", z_myproc, this->numhas, this->has[src], src);*/
  } else {
    if (this->has[src] != -1) this->numhas --;
    this->has[src] = -1;
    /*printf("[%2d]--%d %d %2d\r\n", z_myproc, this->numhas, this->has[src], src);*/
  }
} /* GotSomeNodes */

#endif /* SEQUENTIAL */

void z_spq_init(void)
{
  /* called on every processor by z_init_z_bbp() */
#ifndef SEQUENTIAL
  if (! z_spq_inited) {
    z_spq_inited = 1;
    z_mp_Init();
    someNodeTag = z_InstallMessage(GotSomeNodes);
    emptyTag = z_InstallMessage(GotEmpty);
  }
#endif /* SEQUENTIAL */
} /* z_spq_init */
